#ifdef SIGABRT
scm_c_define ("SIGABRT", scm_from_int (SIGABRT));
#endif
#ifdef SIGALRM
scm_c_define ("SIGALRM", scm_from_int (SIGALRM));
#endif
#ifdef SIGBREAK
scm_c_define ("SIGBREAK", scm_from_int (SIGBREAK));
#endif
#ifdef SIGBUS
scm_c_define ("SIGBUS", scm_from_int (SIGBUS));
#endif
#ifdef SIGCHLD
scm_c_define ("SIGCHLD", scm_from_int (SIGCHLD));
#endif
#ifdef SIGCLD
scm_c_define ("SIGCLD", scm_from_int (SIGCLD));
#endif
#ifdef SIGCONT
scm_c_define ("SIGCONT", scm_from_int (SIGCONT));
#endif
#ifdef SIGEMT
scm_c_define ("SIGEMT", scm_from_int (SIGEMT));
#endif
#ifdef SIGFPE
scm_c_define ("SIGFPE", scm_from_int (SIGFPE));
#endif
#ifdef SIGHUP
scm_c_define ("SIGHUP", scm_from_int (SIGHUP));
#endif
#ifdef SIGILL
scm_c_define ("SIGILL", scm_from_int (SIGILL));
#endif
#ifdef SIGINFO
scm_c_define ("SIGINFO", scm_from_int (SIGINFO));
#endif
#ifdef SIGINT
scm_c_define ("SIGINT", scm_from_int (SIGINT));
#endif
#ifdef SIGIO
scm_c_define ("SIGIO", scm_from_int (SIGIO));
#endif
#ifdef SIGIOT
scm_c_define ("SIGIOT", scm_from_int (SIGIOT));
#endif
#ifdef SIGKILL
scm_c_define ("SIGKILL", scm_from_int (SIGKILL));
#endif
#ifdef SIGLOST
scm_c_define ("SIGLOST", scm_from_int (SIGLOST));
#endif
#ifdef SIGPIPE
scm_c_define ("SIGPIPE", scm_from_int (SIGPIPE));
#endif
#ifdef SIGPOLL
scm_c_define ("SIGPOLL", scm_from_int (SIGPOLL));
#endif
#ifdef SIGPROF
scm_c_define ("SIGPROF", scm_from_int (SIGPROF));
#endif
#ifdef SIGPWR
scm_c_define ("SIGPWR", scm_from_int (SIGPWR));
#endif
#ifdef SIGQUIT
scm_c_define ("SIGQUIT", scm_from_int (SIGQUIT));
#endif
#ifdef SIGRTMAX
scm_c_define ("SIGRTMAX", scm_from_int (SIGRTMAX));
#endif
#ifdef SIGRTMIN
scm_c_define ("SIGRTMIN", scm_from_int (SIGRTMIN));
#endif
#ifdef SIGSEGV
scm_c_define ("SIGSEGV", scm_from_int (SIGSEGV));
#endif
#ifdef SIGSTKFLT
scm_c_define ("SIGSTKFLT", scm_from_int (SIGSTKFLT));
#endif
#ifdef SIGSTKSZ
scm_c_define ("SIGSTKSZ", scm_from_int (SIGSTKSZ));
#endif
#ifdef SIGSTOP
scm_c_define ("SIGSTOP", scm_from_int (SIGSTOP));
#endif
#ifdef SIGSYS
scm_c_define ("SIGSYS", scm_from_int (SIGSYS));
#endif
#ifdef SIGTERM
scm_c_define ("SIGTERM", scm_from_int (SIGTERM));
#endif
#ifdef SIGTRAP
scm_c_define ("SIGTRAP", scm_from_int (SIGTRAP));
#endif
#ifdef SIGTSTP
scm_c_define ("SIGTSTP", scm_from_int (SIGTSTP));
#endif
#ifdef SIGTTIN
scm_c_define ("SIGTTIN", scm_from_int (SIGTTIN));
#endif
#ifdef SIGTTOU
scm_c_define ("SIGTTOU", scm_from_int (SIGTTOU));
#endif
#ifdef SIGUNUSED
scm_c_define ("SIGUNUSED", scm_from_int (SIGUNUSED));
#endif
#ifdef SIGURG
scm_c_define ("SIGURG", scm_from_int (SIGURG));
#endif
#ifdef SIGUSR1
scm_c_define ("SIGUSR1", scm_from_int (SIGUSR1));
#endif
#ifdef SIGUSR2
scm_c_define ("SIGUSR2", scm_from_int (SIGUSR2));
#endif
#ifdef SIGVTALRM
scm_c_define ("SIGVTALRM", scm_from_int (SIGVTALRM));
#endif
#ifdef SIGWINCH
scm_c_define ("SIGWINCH", scm_from_int (SIGWINCH));
#endif
#ifdef SIGXCPU
scm_c_define ("SIGXCPU", scm_from_int (SIGXCPU));
#endif
#ifdef SIGXFSZ
scm_c_define ("SIGXFSZ", scm_from_int (SIGXFSZ));
#endif
